local AceGUI = LibStub("AceGUI-3.0");
local Spreadsheet = LibStub("LibSpreadsheet-1.0");


if ( GnomexcelDB == nil ) then
    GnomexcelDB = {
        books = {}
    };
end

Gnomexcel = {
    mainWin = nil;
    toolBox = nil;
    fileTree = nil;
    revision = tonumber(("$Rev: 013$"):match("(%d+)")) or "0 (dev copy)";
};


local function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

function Gnomexcel:ContextFunction(caller)
    local menu = {
    { text = "Tools:", isTitle = true},
    { text = "Clear cells", func = function() if (Gnomexcel.book) then Gnomexcel:Wipe(); end end },
	{ text = "Copy cells", func = function() if (Gnomexcel.book) then Gnomexcel:Copy(); end end },
    { text = "Cut cells", func = function() if (Gnomexcel.book) then Gnomexcel:Copy(true); end end },
    { text = "Paste cells", disabled = true, func = function() if (Gnomexcel.book) then Gnomexcel:Paste(); end end },
    { text = " ", notClickable = true },
    { text = "Freeze panes here", func = function() if (Gnomexcel.book) then Gnomexcel:Freeze(true); end end },
    { text = "Unfreeze panes", func = function() if (Gnomexcel.book) then Gnomexcel:Freeze(false); end end },
    { text = " ", notClickable = true },
    { text = "Sort rows:", hasArrow = true, menuList = {
        { text = "Ascending  (0..9 / a..z)", func = function() EasyMenu({}, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); Gnomexcel:Sort(Spreadsheet.sortAscending); end },
        { text = "Descending (9..0 / z..a)", func = function() EasyMenu({}, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); Gnomexcel:Sort(Spreadsheet.sortDescending); end }
    }},
    { text = "Cancel" },
    };
    if ( Gnomexcel.bucket ) then
        menu[5].disabled = false;
    end
    EasyMenu(menu, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU");
end

function Gnomexcel:Copy(cut)
    Gnomexcel.bucket = {};
    local r = 0;
    local row = {};
    for key, cell in pairs(Gnomexcel.book.selectedSheet:GetSelectedCells()) do
        if ( r == 0 ) then r = cell.frame.row; end
        if ( cell.frame.row ~= r ) then table.insert(Gnomexcel.bucket , row); row = {}; r = cell.frame.row; end
        local x = deepcopy(cell);
        x.frame = nil;
        x.u = true;
        x.err = nil;
        table.insert(row, x);
    end
    table.insert(Gnomexcel.bucket , row);
    if ( cut ) then
        Gnomexcel:Wipe();
    end
end

function Gnomexcel:Paste()
    local cell = unpack(Gnomexcel.book.selectedSheet:GetSelectedCells(1));
    local pcol, prow = cell.frame.column, cell.frame.row;
    for row, entry in pairs(Gnomexcel.bucket) do
        local rpos = prow + row - 1;
        if ( Gnomexcel.book.selectedSheet.rows[rpos] == nil ) then
            Gnomexcel.book.selectedSheet:AddRow();
        end
        for col, cell in pairs(entry) do
            local cpos = pcol + col - 1;
            if (Gnomexcel.book.selectedSheet.columns[cpos] == nil) then
                Gnomexcel.book.selectedSheet:AddColumn({
                    type="string",
                    name = i,
                    background={0.14,0.14,0.14},
                    color={1,0.95,0.4}
                    }
                );
            end
            Gnomexcel.book.selectedSheet.rows[rpos].u = true;
            local frame = Gnomexcel.book.selectedSheet.rows[rpos].cells[cpos].frame;
            Gnomexcel.book.selectedSheet.rows[rpos].cells[cpos] = deepcopy(cell);
            Gnomexcel.book.selectedSheet.rows[rpos].cells[cpos].frame = frame;
        end
    end
    Gnomexcel.book:Render_Update();
end

function Gnomexcel:PortWindow(what)
    Gnomexcel.portWin:Show();
    Gnomexcel.portGroup:ReleaseChildren();

    if ( what == "export" ) then
        if ( Gnomexcel.book ) then
            local t, r, f = "csv", "all", "formatted";
            local etype = AceGUI:Create("Dropdown");
            etype:SetLabel("Export as:");
            etype:SetList({});
            etype:SetWidth(75);
            etype:AddItem("csv", "CSV");
            etype:AddItem("xml", "XML");
            etype:AddItem("html", "HTML");
            etype:SetValue("csv");
            etype:SetCallback("OnValueChanged", function(key) t = key.value; end);
            Gnomexcel.portGroup:AddChild(etype);

            local range = AceGUI:Create("Dropdown");
            range:SetLabel("Range:");
            range:SetList({});
            range:SetWidth(115);
            range:AddItem("all", "Entire sheet");
            range:AddItem("sel", "Selected cells");
            range:SetValue("all");
            range:SetCallback("OnValueChanged", function(key) r = key.value; end);
            Gnomexcel.portGroup:AddChild(range);

            local format = AceGUI:Create("Dropdown");
            format:SetLabel("Extract:");
            format:SetList({});
            format:SetWidth(215);
            format:AddItem("formatted", "As shown (visual format)");
            format:AddItem("raw", "Raw values (with macro output)");
            format:AddItem("macro", "Raw values (with macro functions)");
            format:SetValue("formatted");
            format:SetCallback("OnValueChanged", function(key) f = key.value; end);
            Gnomexcel.portGroup:AddChild(format);

            local button = AceGUI:Create("Button");
            button:SetText("Export data");
            Gnomexcel.portGroup:AddChild(button);

            local box = AceGUI:Create("MultiLineEditBox");
            box:SetLabel("Data");
            box.width = "fill";
            box:SetHeight(180);
            Gnomexcel.portGroup:AddChild(box);
            button:SetCallback("OnClick", function() Gnomexcel:Export(box, t, r, f); end);

        else
            local text = AceGUI:Create("Label");
            text:SetText("You need to select a book and a sheet before you can export data.");
            Gnomexcel.portGroup:AddChild(text);
        end
    else
        if ( Gnomexcel.book ) then
            local t, r, f = "csv", "new", "formatted";
            local etype = AceGUI:Create("Dropdown");
            etype:SetLabel("Import type:");
            etype:SetList({});
            etype:SetWidth(75);
            etype:AddItem("csv", "CSV");
       --     etype:AddItem("xml", "XML");
       --     etype:AddItem("html", "HTML");
            etype:SetValue("csv");
            etype:SetCallback("OnValueChanged", function(key) t = key.value; end);
            Gnomexcel.portGroup:AddChild(etype);

            local range = AceGUI:Create("Dropdown");
            range:SetLabel("Import to:");
            range:SetList({});
            range:SetWidth(115);
            range:AddItem("new", "New sheet");
            range:AddItem("cur", "Current sheet");
            range:SetValue("new");
            range:SetCallback("OnValueChanged", function(key) r = key.value; end);
            Gnomexcel.portGroup:AddChild(range);

            local button = AceGUI:Create("Button");
            button:SetText("Import data");
            Gnomexcel.portGroup:AddChild(button);

            local box = AceGUI:Create("MultiLineEditBox");
            box:SetLabel("Data");
            box.width = "fill";
            box:SetHeight(180);
            Gnomexcel.portGroup:AddChild(box);
            button:SetCallback("OnClick", function() Gnomexcel:Import(box, t, r, f); end);
        else
            local text = AceGUI:Create("Label");
            text:SetText("You need to select a book and a sheet before you can export data.");
            Gnomexcel.portGroup:AddChild(text);
        end
    end
end

function Gnomexcel:Import(box, etype, edestination)
    local lines = {};
    local delims = {};
    for line in box:GetText():gmatch("([^\n]*)(\n?)") do
        table.insert(lines, line);
        for char in line:gmatch("([%p%c])") do
            delims[char] = (delims[char] or 0) + 1;
        end
    end
    local sep = ",";
    local x = 1;
    for k, v in pairs(delims) do
        if ( v > x ) then x = v; sep = k; end
    end
    local rows = {};
    for k, line in pairs(lines) do
       line = line .. sep;
       local t = {}
       local fieldstart = 1
       repeat
          if string.find(line, '^"', fieldstart) then
             local a, c
             local i = fieldstart
             repeat
                a, i, c = string.find(line, '"("?)', i+1)
             until c ~= '"'
             if not i then error('unmatched "') end
             local f = string.sub(line, fieldstart+1, i-1)
             table.insert(t, {data = (string.gsub(f, '""', '"'))})
             fieldstart = string.find(line, sep, i) + 1
          else
             local nexti = string.find(line, sep, fieldstart)
             table.insert(t, {data=string.sub(line, fieldstart, nexti-1)})
             fieldstart = nexti + 1
          end
       until fieldstart > string.len(line)
       table.insert(rows, t);
    end
    if ( edestination == "new" ) then
        local sheet = Gnomexcel.book:Add(nil);
        sheet.background = {1,1,1,0.4};
        Gnomexcel.book.selectedSheet:AddColumn({
                    type="string",
                    name = i,
                    background={0.14,0.14,0.14},
                    color={1,0.95,0.4},
                    u = true
                    }
                );
        Gnomexcel.book.selectedSheet:AddRow();
        sheet.isInteractive = true;
        Gnomexcel.book:Render();
    end
    for key, row in pairs(rows) do
        if ( Gnomexcel.book.selectedSheet.rows[key] == nil ) then
            Gnomexcel.book.selectedSheet:AddRow();
        end
        for col, cell in pairs(row) do
            if (Gnomexcel.book.selectedSheet.columns[col] == nil) then
                Gnomexcel.book.selectedSheet:AddColumn({
                    type="string",
                    name = i,
                    background={0.14,0.14,0.14},
                    color={1,0.95,0.4},
                    u = true
                    }
                );
            end
            if ( Gnomexcel.book.selectedSheet.rows[key] and Gnomexcel.book.selectedSheet.rows[key].cells[col] ) then
                Gnomexcel.book.selectedSheet.rows[key].cells[col].data = cell.data;
                Gnomexcel.book.selectedSheet.rows[key].cells[col].u = true;
            end
        end
    end
    Gnomexcel.book:Render_Update();
end

function Gnomexcel:Export(box, etype, erange, eformat)
    local cols = {};
    local rows = {};
    if ( erange == "sel" ) then
        cols = Gnomexcel.book.selectedSheet:GetSelectedCols();
        rows = Gnomexcel.book.selectedSheet:GetSelectedRows();
    else
        for key, col in pairs (Gnomexcel.book.selectedSheet.columns) do
            table.insert(cols, key);
        end
        for key, row in pairs (Gnomexcel.book.selectedSheet.rows) do
            table.insert(rows, key);
        end
    end
    local output = "";

    if ( etype == "csv" ) then
        for rPos, row in pairs(rows) do
            local r = {};
            for index, col in pairs(cols) do
                local cell = Gnomexcel.book.selectedSheet.rows[row].cells[col];
                if ( eformat == "formatted" ) then
                    table.insert(r, [["]] .. (cell.text or ""):gsub([["]], [[""]]) .. [["]]);
                elseif ( eformat == "raw" ) then
                    local data = (cell.MathData or cell.data or nil);
                    if ( type(data) == "string" ) then data = [["]] .. data:gsub([["]], [[""]]) .. [["]]; end
                    table.insert(r, data);
                else
                    local data = (cell.data or nil);
                    if ( type(data) == "string" ) then data = [["]] .. data:gsub([["]], [[""]]) .. [["]]; end
                    table.insert(r, data);
                end
            end
            output = output .. table.concat(r, ",") .. "\n";
        end
    end
    if ( etype == "xml" ) then
        output = "<xml>\n    <table>\n"
        for rPos, row in pairs(rows) do
            local r = {};
            output = output .. ("        <row id=\"%u\">\n"):format(row);
            for index, col in pairs(cols) do
                output = output .. ("            <cell id=\"%s%u\">"):format(Spreadsheet.cols[col], row);
                local cell = Gnomexcel.book.selectedSheet.rows[row].cells[col];
                if ( eformat == "formatted" ) then
                    output = output .. (cell.text or ""):gsub("<", "&lt;"):gsub(">", "&gt;");
                elseif ( eformat == "raw" ) then
                    local data = ("%s"):format(cell.MathData or cell.data or ""):gsub("<", "&lt;"):gsub(">", "&gt;");
                    output = output .. data;
                else
                    local data = ("%s"):format(cell.data or ""):gsub("<", "&lt;"):gsub(">", "&gt;");
                    output = output .. data:gsub("<", "&lt;");
                end
                output = output .. "</cell>\n";
            end
            output = output .. "        </row>\n";
        end
        output = output .. "    </table>\n</xml>";
    end
    if ( etype == "html" ) then
        output = "<html><body>\n    <table>\n"
        for rPos, row in pairs(rows) do
            local r = {};
            output = output .. ("        <tr id=\"%u\">\n"):format(row);
            for index, col in pairs(cols) do
                output = output .. ("            <td id=\"%s%u\">"):format(Spreadsheet.cols[col], row);
                local cell = Gnomexcel.book.selectedSheet.rows[row].cells[col];
                if ( eformat == "formatted" ) then
                    output = output .. (cell.text or ""):gsub("<", "&lt;"):gsub(">", "&gt;");
                elseif ( eformat == "raw" ) then
                    local data = ("%s"):format(cell.MathData or cell.data or ""):gsub("<", "&lt;"):gsub(">", "&gt;");
                    output = output .. data;
                else
                    local data = ("%s"):format(cell.data or ""):gsub("<", "&lt;"):gsub(">", "&gt;");
                    output = output .. data:gsub("<", "&lt;");
                end
                output = output .. "</td>\n";
            end
            output = output .. "        </tr>\n";
        end
        output = output .. "    </table>\n</body></xml>";
    end
    box:SetText(output);
end

function Gnomexcel:Freeze(toggle)
    local c, r = 0, 0;
    local cell = unpack(Gnomexcel.book.selectedSheet:GetSelectedCells(1));
    if ( cell ) then
        c = cell.l[1] - 1;
        r = cell.l[2] - 1;
    end
    Gnomexcel.book.selectedSheet:Freeze(toggle, c, r);
    Gnomexcel.book:Render_Real();
end

function Gnomexcel:Sort(direction)
    local tbl = {};
    local col = unpack(Gnomexcel.book.selectedSheet:GetSelectedCols(1));
    if ( col ) then
        table.insert(tbl, col);
        table.insert(tbl, direction);
        local heap = {};
        local ref = Gnomexcel.book.selectedSheet.rows;
        for k, row in pairs(Gnomexcel.book.selectedSheet:GetSelectedRows(1)) do
            table.insert(heap, Gnomexcel.book.selectedSheet.rows[row]);
        end
        Gnomexcel.book.selectedSheet.rows = heap;
        Gnomexcel.book.selectedSheet:SortByColumn(unpack(tbl));
        Gnomexcel.book.selectedSheet.rows = ref;
        for k, row in pairs(Gnomexcel.book.selectedSheet:GetSelectedRows(1)) do
            Gnomexcel.book.selectedSheet.rows[row] = heap[k];
        end
        heap = {};
    end
    Gnomexcel.book:Render_Update();
end


function Gnomexcel:AddRow(num)
    local start = Gnomexcel.book.selectedSheet.selectStartRow or false;
    if ( start ) then
        for i = 1, num do
            Gnomexcel.book.selectedSheet:AddRowAt(start, "","","","","","","","","","","","","","","","","","","","","","","","","","");
        end
    else
        for i = 1, num do
            Gnomexcel.book.selectedSheet:AddRow("","","","","","","","","","","","","","","","","","","","","","","","","","");
        end
    end
    Gnomexcel.book:Render_Update();
end

function Gnomexcel:AddColumn(num)
    local start = Gnomexcel.book.selectedSheet.selectStartCol or false;
    if ( start ) then
        for i = 1, num do
            Gnomexcel.book.selectedSheet:AddColumnAt(start,
        {
            type="string",
            name = i,
            background={0.14,0.14,0.14},
            color={1,0.95,0.4}
        }
        );
        end
    else
        for i = 1, num do
            Gnomexcel.book.selectedSheet:AddColumn(
        {
            type="string",
            name = i,
            background={0.14,0.14,0.14},
            color={1,0.95,0.4}
        }
        );
        end
    end
    Gnomexcel.book:Render_Update();
end

function Gnomexcel:SetStat(prop, val)
    if ( Gnomexcel.book ) then
        if ( val ~= "void" ) then
            for k, cell in pairs(Gnomexcel.book.selectedSheet:GetSelectedCells()) do
                cell[prop] = val;
                cell.u = true;
            end
        end
        Gnomexcel.book:Render_Real();
    end
end

function Gnomexcel:Wipe()
    for k, cell in pairs(Gnomexcel.book.selectedSheet:GetSelectedCells()) do
        cell.text = "";
        cell.data = nil;
        cell.MathData = nil;
        cell.MathFuntion = nil;
    end
    Gnomexcel.book:Render_Update();
end

function Gnomexcel:DeleteRows()
    for k, row in pairs(Gnomexcel.book.selectedSheet:GetSelectedRows()) do
        Gnomexcel.book.selectedSheet:DeleteRow(row);
    end
    Gnomexcel.book:Render_Update();
end

function Gnomexcel:DeleteCols()
    local x = 0;
    for k, col in pairs(Gnomexcel.book.selectedSheet:GetSelectedCols()) do
        Gnomexcel.book.selectedSheet:DeleteColumn(col+x);
        x = x - 1;
    end
    Gnomexcel.book:Render_Update();
end


function Gnomexcel:CreateBook(title)
    local book = Spreadsheet:Book(title);
    book:StrictMode(true);
    local sheet = book:Add("Sheet 1");
    sheet.background = {1,1,1,0.4};

    for i = 1, 15 do
        sheet:AddColumn(
        {
            type="string",
            name = i,
            background={0.14,0.14,0.14},
            color={1,0.95,0.4}
        }
        );
    end
    for i = 1, 15 do
        sheet:AddRow("","","","","","","","","","","","","","","","","","","","","","","","","","");
    end
    sheet.isInteractive = true;
    book:EnableMaths(true);
    local hash = string.format("%8x-%8x-%8x", math.random(time()), math.random(time()/3), math.random(time()-400));
    book.hash = hash;

    table.insert(GnomexcelDB.books,
        {
        author = GetUnitName("Player"),
        realm = GetRealmName(),
        created = time(),
        file = book,
        hash = hash;
        }
    );
end

function Gnomexcel:DeleteBook()
    if ( Gnomexcel.book ~= nil ) then
        Gnomexcel.book:Hide();
        Gnomexcel.book:Close();
    end
    for key,book in pairs(GnomexcelDB.books) do
        if ( book.hash == Gnomexcel.book.hash ) then
            table.remove(GnomexcelDB.books, key);
            break;
        end
    end
    Gnomexcel:GetBooks();
end

function Gnomexcel:GetBooks()
    local tbl = {
        {
            value = "$$Function=NewBook$$",
            text = "Create a new book",
            icon = "Interface\\Icons\\INV_Misc_Book_06",
        },
        {
            value = "$$Empty$$",
            text = "   ",
            icon = "",
            disabled = true,
        },
        {
            value = "$$Empty$$",
            text = "|cffffffffBooks in your library:|r",
            icon = "",
            disabled = true,
        },

    };
    for key, book in pairs(GnomexcelDB.books) do
        table.insert(tbl, {
            value = "$$Book=" .. key .. "$$",
            text = book.file.title,
            icon = ("Interface\\Icons\\INV_Misc_Book_%02u"):format(math.random(0, 4) + 7);
        }
        );
    end
    Gnomexcel.fileTree:SetTree(tbl);
    Gnomexcel.fileTree:SetUserData("tree", tbl);

    Gnomexcel.fileTree:SetCallback("OnClick",
        function(caller, event, value)
            local subtree, subval = (value):match("%$(%a+)=([^%$]+)%$");
            if ( subtree == "Function" ) then
                if ( subval == "NewBook" ) then
                    local frame = AceGUI:Create("Frame");
                    frame:SetWidth(300);
                    frame:SetHeight(100);
                    frame:SetTitle("Create new book");
                    local box = AceGUI:Create("EditBox");
                    box:SetText("Enter the name of the book");
                    frame:AddChild(box);
                    frame:Show();
                    box:SetCallback("OnEnterPressed",
                        function(caller, event, title)
                            Gnomexcel:CreateBook(title);
                            Gnomexcel:GetBooks();
                            frame:ReleaseChildren();
                            frame:Release();
                        end
                    );
                end
            end
            if ( subtree == "Book" ) then
                if ( Gnomexcel.book ~= nil and Gnomexcel.book.Hide ~= nil ) then
                    Gnomexcel.book:Hide();
                    Gnomexcel.book:Close();
                end
                Gnomexcel.book = Spreadsheet:Load(GnomexcelDB.books[tonumber(subval)].file);
                Gnomexcel.book:SetStyle("sheet");
                Gnomexcel.book:ShowSelector(true);
                Gnomexcel.book:Select();
                Gnomexcel.book:SetContextFunction(Gnomexcel.ContextFunction);
                frame = Gnomexcel.book:Render();
                Gnomexcel.container:ReleaseChildren();
                frame:SetParent(Gnomexcel.container.frame);
                frame:SetPoint("TOPLEFT", Gnomexcel.container.frame, "TOPLEFT", 10, -10);
                frame:SetPoint("BOTTOMRIGHT", Gnomexcel.container.frame, "BOTTOMRIGHT", -10, 10);
                frame:Show();
            end
        end
    );
    Gnomexcel.fileTree:SetCallback("OnButtonEnter",
        function(caller, event, value)
            local subtree, subval = (value):match("%$(%a+)=([^%$]+)%$");
            if ( subtree == "Book" ) then
                if ( tonumber(subval) ~= 0 ) then
                    local book = GnomexcelDB.books[tonumber(subval)];
                    GameTooltip:SetOwner(caller.frame, "ANCHOR_TOPLEFT");
                    GameTooltip:ClearLines();
                    GameTooltip:SetText(book.file.title .. ":");
                    GameTooltip:AddDoubleLine("|cff55DD11Author:|r", book.author .. " of " .. book.realm);
                    GameTooltip:AddDoubleLine("|cff55DD11Created:|r", date(nil, book.created));
                    GameTooltip:Show();
                end
            end
        end);
    Gnomexcel.fileTree:SetCallback("OnButtonLeave", function() GameTooltip:Hide(); end);
    Gnomexcel.fileTree:EnableButtonTooltips(false);
end

function Gnomexcel:MakeToolbox()

    local group = AceGUI:Create("SimpleGroup");
    group:SetWidth(420);
    group:SetLayout("Flow");

    local icon = AceGUI:Create("Icon");
    icon:SetWidth(48);
    icon:SetImage([[Interface\Icons\INV_Misc_Bomb_07]]);
    icon:SetImageSize(32,32);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Delete book"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetCallback("OnClick", function() if ( Gnomexcel.book ) then Gnomexcel:DeleteBook() end end );
    group:AddChild(icon);

    icon = AceGUI:Create("Icon");

    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Add sheet"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\INV_Misc_Bandage_10]]);
    icon:SetCallback("OnClick", function()
    if ( Gnomexcel.book ) then
        local sheet = Gnomexcel.book:Add(nil);
        sheet.background = {1,1,1,0.4};

    for i = 1, 15 do
        sheet:AddColumn(
        {
            type="string",
            name = i,
            background={0.14,0.14,0.14},
            color={1,0.95,0.4}
        }
        );
    end
    for i = 1, 15 do
        sheet:AddRow("","","","","","","","","","","","","","","","","","","","","","","","","","");
    end
    sheet.isInteractive = true;
        Gnomexcel.book:Render(); end end);
    group:AddChild(icon);




    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Delete sheet"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\INV_Misc_Map_01]]);
    icon:SetCallback("OnClick", function() if ( Gnomexcel.book ) then Gnomexcel.book:Delete(nil); Gnomexcel.book:Render(); end end);
    group:AddChild(icon);



    local menu = {
    { text = "Add rows/columns:", isTitle = true},
    { text = "Add a single column", func = function() Gnomexcel:AddColumn(1) end },
    { text = "Add 5 columns", func = function() Gnomexcel:AddColumn(5) end },
    { text = "Add 10 columns", func = function() Gnomexcel:AddColumn(10) end },
    { text = " "},
    { text = "Add a single row", func = function() Gnomexcel:AddRow(1) end },
    { text = "Add 5 rows", func = function() Gnomexcel:AddRow(5) end },
    { text = "Add 10 rows", func = function() Gnomexcel:AddRow(10) end },
    { text = "Add 20 rows", func = function() Gnomexcel:AddRow(20) end },
    { text = "Add 50 rows", func = function() Gnomexcel:AddRow(50) end },
    { text = "",},
    { text ="Close menu" }
    };

    icon = AceGUI:Create("Icon");
    icon:SetImage([[Interface\Icons\Spell_ChargePositive]]);
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Add rows/columns"); GameTooltip:AddLine("|cff55EE22Tip: If you have a row, column or cell selected, any added |nrows/columns will be inserted BEFORE that selection.|r"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetCallback("OnClick", function() EasyMenu(menu, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); end);

    group:AddChild(icon);


    local menu = {
    { text = "Delete rows/columns:", isTitle = true},
    { text = "Delete row(s)", func = function() Gnomexcel:DeleteRows() end },
    { text = "Delete column(s)", func = function() Gnomexcel:DeleteCols() end },
    { text = "",},
    { text ="Close menu" }
    };
    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Delete rows/columns"); GameTooltip:AddLine("|cff55EE22Tip: Any rows or columns following the selected|n ones will be shifted up or to the left upon deleting.|nIf you just wish to clear the cells,|nselect them and use the [eraser] button.|r");GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\Spell_ChargeNegative]]);
    icon:SetCallback("OnClick", function() EasyMenu(menu, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); end);
    group:AddChild(icon);

--[[
    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Send book"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([ [Interface\Icons\Inv_Letter_15] ]);
    icon:SetCallback("OnClick", function() if ( Gnomexcel.book ) then Gnomexcel.book.selectedSheet:icon(""); Gnomexcel.book:Render_ScrollFunction(); end end);
    group:AddChild(icon);
]]
    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Import/Export data"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\INV_Misc_PunchCards_Yellow]]);
    icon:SetCallback("OnClick", function() Gnomexcel.portGroup:SelectTab("export"); end);
    group:AddChild(icon);





--[[
    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Settings"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([ [Interface\Icons\INV_Misc_EngGizmos_30] ]);
    --icon:SetCallback("OnClick", function() EasyMenu(menu, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); end);
    group:AddChild(icon);
]]
    Gnomexcel.toolBox:AddChild(group);



    group = AceGUI:Create("SimpleGroup");
    group:SetWidth(230);
    group:SetLayout("Flow");

    local menu = {
    { text = "Data type:", isTitle = true},
    { text = "Numbers", func = function() Gnomexcel:SetStat("t", "number"); end },
    { text = "Text", func = function()  Gnomexcel:SetStat("t", "string"); end },
    { text = "WoW Currency", func = function()  Gnomexcel:SetStat("t", "currency"); end },
    { text = "WoW Item/Spell", func = function()  Gnomexcel:SetStat("t", "link"); end },
    { text = "Booleans", func = function()  Gnomexcel:SetStat("t", "boolean"); end },
    };
    --[[
    INV_Misc_Spyglass_02



http://wowprogramming.com/BlizzArt/Icons/INV_Holiday_Christmas_Wrapping_01.png
http://wowprogramming.com/BlizzArt/Icons/INV_Misc_QuestionMark.png
]]


    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Data type"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\INV_Misc_QuestionMark]]);
    icon:SetCallback("OnClick", function() EasyMenu(menu, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); end);
    group:AddChild(icon);



    local menu = {
    { text = "Data precision:", isTitle = true},
    { text = "Whole numbers (12345)", func = function() Gnomexcel:SetStat("d", 0); end },
    { text = "1 decimal  (1234.5)", func = function() Gnomexcel:SetStat("d", 1); end },
    { text = "2 decimals (123.45)", func = function()  Gnomexcel:SetStat("d", 2); end },
    { text = "3 decimals (12.345)", func = function()  Gnomexcel:SetStat("d", 3); end },
    { text = "4 decimals (1.2345)", func = function()  Gnomexcel:SetStat("d", 4); end },
    };
    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Data precision"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\INV_Misc_Spyglass_02]]);
    icon:SetCallback("OnClick", function() EasyMenu(menu, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); end);
    group:AddChild(icon);


    local menu = {
    { text = "Prefix/Suffix:", isTitle = true},
    { text = "(None)", func = function() Gnomexcel:SetStat("s", ""); Gnomexcel:SetStat("p", ""); end },
    { text = "% (Percent)", func = function() Gnomexcel:SetStat("s", "%");Gnomexcel:SetStat("p", ""); end },
    { text = "mp5 (Mana per 5 seconds)", func = function()  Gnomexcel:SetStat("p", ""); Gnomexcel:SetStat("s", " mp5"); end },
    { text = "dps (Damage per second)", func = function()  Gnomexcel:SetStat("p", ""); Gnomexcel:SetStat("s", " dps"); end },
    { text = "hps (Healing per second)", func = function()  Gnomexcel:SetStat("p", ""); Gnomexcel:SetStat("s", " hps"); end },
    { text = "tps (Threat per second)", func = function()  Gnomexcel:SetStat("p", ""); Gnomexcel:SetStat("s", " tps"); end },
    };
    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Suffix/Prefix"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\INV_Misc_Ribbon_01]]);
    icon:SetCallback("OnClick", function() EasyMenu(menu, Gnomexcel.menuFrame, "cursor", 0 , 0, "MENU"); end);
    group:AddChild(icon);

    icon = AceGUI:Create("Icon");
    icon:SetImageSize(32,32);
    icon:SetWidth(48);
    icon:SetCallback("OnEnter", function(caller) GameTooltip:ClearLines(); GameTooltip:SetOwner(caller.frame, "ANCHOR_BOTTOM"); GameTooltip:SetText("Erase selected cells"); GameTooltip:Show(); end);
    icon:SetCallback("OnLeave", function() GameTooltip:Hide(); end);
    icon:SetImage([[Interface\Icons\INV_Inscription_InkWhite01]]);
    icon:SetCallback("OnClick", function() Gnomexcel:Wipe(); end);
    group:AddChild(icon);


    local textColor = AceGUI:Create("ColorPicker");
    textColor:SetLabel("Text");
    textColor:SetWidth(60);
    textColor:SetColor(1,0.95,0.4);
    textColor:SetCallback("OnValueConfirmed",
        function(caller, event, r, g, b)
            if ( Gnomexcel.book ) then
                for k, cell in pairs(Gnomexcel.book.selectedSheet:GetSelectedCells()) do
                    cell.c = {r,g,b,1};
                    cell.u = true;
                end
                Gnomexcel.book:Render_Real();
            end
        end
    );
    group:AddChild(textColor);

    local background = AceGUI:Create("ColorPicker");
    background:SetLabel("Background");
    background:SetColor(0.14,0.14,0.14);
    background:SetWidth(89);
    background:SetCallback("OnValueConfirmed",
        function(caller, event, r, g, b)
            if ( Gnomexcel.book ) then
                for k, cell in pairs(Gnomexcel.book.selectedSheet:GetSelectedCells()) do
                    cell.b = {r,g,b,1};
                    cell.u = true;
                end
                Gnomexcel.book:Render_Real();
            end
        end
    );
    group:AddChild(background);

    Gnomexcel.toolBox:AddChild(group);
end


function Gnomexcel:MainWindow()
    local f = AceGUI:Create("Frame")
    f:SetCallback("OnClose",function(widget) f:Hide(); end)
    f:SetTitle("Gnomexcel rev. " .. Gnomexcel.revision);
    f:SetStatusText("Welcome!");
    f:SetLayout("Flow")
    f:SetWidth(730);
    f:SetAutoAdjustHeight(false);
    f:SetHeight(480);

    local toolBox = AceGUI:Create("SimpleGroup");
    toolBox.width= "fill";
    toolBox:SetHeight(80);
    toolBox:SetLayout("Flow")
    toolBox:SetAutoAdjustHeight(false);
    Gnomexcel.toolBox = toolBox;
    f:AddChild(toolBox);

    local fileTree = AceGUI:Create("TreeGroup");
    fileTree.width = "fill";
    fileTree.height = "fill";

    f:AddChild(fileTree);

    local container = AceGUI:Create("SimpleGroup");

    fileTree:SetLayout("fill");
    fileTree:AddChild(container);

    f.frame:SetScript("OnShow", function() Gnomexcel:GetBooks(); end);
    Gnomexcel.mainWin = f;
    Gnomexcel.container = container;
    f:Hide();
    Gnomexcel.book = book;
    Gnomexcel.fileTree = fileTree;
    Gnomexcel:MakeToolbox();

    local t = AceGUI:Create("Label");
    t:SetText("Welcome to Gnomexcel! To get started, select one of your previous books or create a new one to start making spreadsheets!");
    t:SetWidth(400);
    container:AddChild(t);

    Gnomexcel.menuFrame = CreateFrame("Frame", "GnomexcelMenu", UIParent, "UIDropDownMenuTemplate")

    Gnomexcel.portWin = AceGUI:Create("Frame");
    Gnomexcel.portWin:SetWidth(510);
    Gnomexcel.portWin:SetHeight(375);
    Gnomexcel.portWin:SetLayout("Fill");
    Gnomexcel.portWin:SetTitle("Data Import/export");

    Gnomexcel.portGroup = AceGUI:Create("TabGroup");
    Gnomexcel.portGroup:SetLayout("Flow");
    Gnomexcel.portGroup:SetAutoAdjustHeight(false);
    Gnomexcel.portGroup:SetTabs({
            [1] = { value = "export", text = "Export data" },
            [2] = { value = "import", text = "Import data" }
        });
    Gnomexcel.portGroup:SelectTab("export");
    Gnomexcel.portGroup:SetCallback("OnGroupSelected", function(group, event, id) Gnomexcel:PortWindow(id); end);
    Gnomexcel.portWin:AddChild(Gnomexcel.portGroup);
    Gnomexcel.portWin:Hide();

    local eventFrame = CreateFrame("Frame", nil);
    eventFrame:RegisterEvent("PLAYER_LOGOUT");
    eventFrame:SetScript("OnEvent", function(caller, event) print(caller, event); if ( event == "PLAYER_LOGOUT" ) then if ( Gnomexcel.book ) then Gnomexcel.book:Close(); end end end);
    eventFrame:SetParent(UIParent);
    eventFrame:Show();
end
do
    Gnomexcel:MainWindow();
    SLASH_GNOMEXCEL1, SLASH_GNOMEXCEL2 = '/gxl', '/gnomexcel';
function SlashCmdList.GNOMEXCEL(msg, editbox)
 Gnomexcel.mainWin:Show();
end

end
