local Grapple = _G.Grapple
local Graph = LibStub:GetLibrary("LibGraph-2.0")
local AceGUI = LibStub("AceGUI-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale("Grapple");
local Spreadsheet = LibStub("LibSpreadsheet-1.0");
local wipe,remove,sort = table.wipe,table.remove,table.sort
local match,find,gmatch,sub = string.match,string.find,string.gmatch,string.sub
local _G,select,tostring,type,tonumber = _G,select,tostring,type,tonumber
local UnitName,UnitGUID,UnitClass,UnitAffectingCombat,UnitHealth,UnitHealthMax,UnitPower,UnitMaxPower,UnitIsDeadOrGhost = 
		UnitName,UnitGUID,UnitClass,UnitAffectingCombat,UnitHealth,UnitHealthMax,UnitPower,UnitMaxPower,UnitIsDeadOrGhost

local function deepcopy(object)
	local lookup_table = {}
	local function _copy(object)
		if type(object) ~= "table" then
			return object
		elseif lookup_table[object] then
			return lookup_table[object]
		end
		local new_table = {}
		lookup_table[object] = new_table
		for index, value in pairs(object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(object))
	end
	return _copy(object)
end

Grapple.SortFunc = function (a, b)
	if (a[Grapple.sortFunc] or 0) > (b[Grapple.sortFunc] or 0) then
		return true
	elseif (a[Grapple.sortFunc] or 0) == (b[Grapple.sortFunc] or 0) then
		if (a["name"] or "") < (b["name"] or "")  then
			return true
		end
	end
	return false
end

function Grapple:MainWindow()
	
	Grapple.currentTab = "nada";
	-- Create a container frame
	local f = AceGUI:Create("Frame")
	f:SetCallback("OnClose",function(widget) f:Hide(); end)
	f:SetTitle("Grapple (rev. " .. Grapple.revision .. ")");
	f:SetStatusText(L["Status: Not in raid combat, monitoring disabled."]);
	f:SetLayout("Flow")
	f:SetWidth(620);
	f:SetAutoAdjustHeight(false);
	f:SetHeight(590);
	
	local drop = AceGUI:Create("Dropdown");
	drop:SetWidth(250);
	drop:SetLabel(L["Select a fight:"]);
	drop.Update = function()
		local tbl = {
			[-1] = L["Entire Session"],
			[0] =  L["Current Fight"],
		};
		for k, v in pairs (GrappleDB.fightHistory) do
			table.insert(tbl, v.title);
		end
		drop:SetList(tbl);
		drop:SetValue(GrappleDB.fightShown);
	end;
		
	drop.Update();
	drop:SetCallback("OnValueChanged", function(key) GrappleDB.fightShown = key.value; Grapple:PopulateWindow(); if ( Grapple.currentTab == "spells" ) then Grapple:ShowSpells(); end end);
	f:AddChild(drop);
	Grapple.fightList = drop;
	
	local label = AceGUI:Create("Label");
	label:SetText("");
	label:SetWidth(10);
	f:AddChild(label);
	
	local btn = AceGUI:Create("Button");
	btn:SetText(L["Delete entry"]);
	btn:SetWidth(110);
	btn:SetCallback("OnClick", 
		function()
			local n = GrappleDB.fightShown;
			if ( n > 0 ) then 
				table.remove(GrappleDB.fightHistory, n);
				Grapple:ShowDetails(nil);
				GrappleDB.fightShown = 0;
				Grapple.fightList:Update();
				Grapple:PopulateWindow();
			end
			end
		);
	f:AddChild(btn);
	
	local btn = AceGUI:Create("Button");
	btn:SetText(L["Clear log"]);
	btn:SetWidth(95);
	btn:SetCallback("OnClick", function() Grapple:Print(L["Fight data has been cleared."]); Grapple:ResetFights();Grapple:ShowDetails(nil); end)
	f:AddChild(btn);
	
	local btn = AceGUI:Create("Button");
	btn:SetText(L["Fix combat log"]);
	btn:SetWidth(115);
	btn:SetCallback("OnClick", function() Grapple:Print(L["Combat log API has been restarted."]); CombatLogClearEntries(); end)
	f:AddChild(btn);
    Grapple.fumbulator = Spreadsheet:Book("Grapple Fumbulator");
    Grapple.fumbulator:StrictMode(true);
    Grapple.fumbulator:Add("Debuffs"):AddColumn(
        {name=L["Pos"], type="number", sort=true},
        {name=L["Character"], sort=true},
        {name=L["Applications"], type="number", dominant=true, sort=true},
        {name=L["Stack count"], sort=true}
    );
    Grapple.fumbulator:Add("Aggro"):AddColumn(
        {name=L["Pos"], type="number"},
        {name=L["Character"], sort=true},
        {name=L["About to aggro"], type="number", suffix=" sec", dominant=true},
        {name=L["Losing aggro"], type="number", suffix=" sec"}
    );
    
    Grapple.scoreboard = Spreadsheet:Book("Grapple Scoreboard");
	local sheet = Grapple.scoreboard:Add("Scores");
    sheet.cellSpacing = 1;
    sheet.cellPadding = 2;
    sheet:AddColumn(
        {name=L["Character"], type="string", interactive=false, onClick=function(id) Grapple.tabGroup:SelectTab("graph"); Grapple:ShowDetails(id[1], id[2]); end, sort=true},
        {name=L["Build"], type="string"},
        {name=L["Uptime"], type="number", suffix="%", minval=0, maxval=100, decimals=2, sort=true},
        {name=L["Survival"], type="number", decimals=2, sort=true},
        {name=L["Activity"], type="number", decimals=2, suffix=" ApM", sort=true},
        {name=L["Fumbles"], type="string", sort=true},
        {name=L["Focus"], type="number", decimals=2, sort=true},
        {name=L["GrappleScore"], type="number", minval=5, maxval=750, decimals=2, sort=true}
    );
    sheet:SetColorRange("H", {1,0,0},{1,1,0},{0,1,0});
	
	
	Grapple.mainWin = f;
	Grapple.mainWin:Hide();
	Grapple:PopulateWindow();
	group = AceGUI:Create("TabGroup");
	group:SetLayout("Flow");
	group:SetWidth(620);
	group:SetHeight(235);
	group:SetTabs({
			[1] = { value = "graph", text = L["Player details"] },
			[2] = { value = "report", text = L["Broadcast data"] },
			[3] = { value = "spells", text = L["Fumbulator™"] },
			[4] = { value = "options", text = L["Internals"] },
			[5] = { value = "about", text = L["About"] },
		});
	group:SelectTab("graph");
	group:SetAutoAdjustHeight(false);
	Grapple.mainWin:AddChild(group);
	Grapple.tabGroup = group;
	Grapple.tabGroup:SetCallback("OnGroupSelected", function(group, event, id) Grapple.tabGroup:ReleaseChildren(); Grapple.currentTab = id; if ( id == "graph" ) then Grapple:ShowDetails(nil); elseif ( id == "report") then Grapple:ShowReport(); elseif ( id == "about" ) then Grapple:ShowAbout(); elseif ( id == "spells" ) then Grapple:ShowSpells(); else Grapple:ShowOptions(); end end);
	Grapple.tabGroup:SelectTab("graph");	
end

function Grapple:ChangeBuild(who)
	if ( who ~= nil ) then
		local char = GrappleDB.characters:Fetch(who.guid);
		if ( char ~= nil ) then
			who.build = who.build + 1;
			if ( Grapple.scoreModifiers[char.class][who.build] == nil ) then
				who.build = 0;
			end
		end
	end
	
	Grapple:SilentPopulate(GrappleDB.fightHistory[GrappleDB.fightShown]);
	Grapple:PopulateWindow();
end



function Grapple:ShowReport()
	if ( Grapple.graph ~= nil ) then
		Grapple.graph:Hide();
	end
	if ( GrappleDB.recipient == nil ) then
		GrappleDB.recipient = 2;
	end
	drop = AceGUI:Create("Dropdown");
	drop:SetWidth(140);
	drop:SetLabel(L["Broadcast channel:"]);
	drop:SetCallback("OnValueChanged", function(key) GrappleDB.recipient = key.value; end);
	local tbl = deepcopy(Grapple.reportList);
	for n = 1, 9 do
		local id, name = GetChannelName(n);
		if ( id ~= 0 ) then
			tbl[id + 10] = id .. ". " .. name;
		end
	end
	drop:SetList(tbl);
	drop:SetText(tbl[GrappleDB.recipient]);
	Grapple.tabGroup:AddChild(drop);
	
	drop = AceGUI:Create("Dropdown");
	drop:SetWidth(100);
	drop:SetLabel(L["Max no. of entries:"]);
	local tbl = {};
	for i = 1, 25 do
		table.insert(tbl, i);
	end
	table.insert(tbl, 40);
	GrappleDB.recipientLimit = 25;
	drop:SetList(tbl);
	drop:SetText("25");
	drop:SetCallback("OnValueChanged", function(key) GrappleDB.recipientLimit = key.value; end);
	Grapple.tabGroup:AddChild(drop);
	
	local btn = AceGUI:Create("Button");
	btn:SetText(L["Send report"]);
	btn:SetWidth(100);
	btn:SetCallback("OnClick", function() Grapple:Report(Grapple.reportList[GrappleDB.recipient]); end)
	Grapple.tabGroup:AddChild(btn);
end


function Grapple:ShowSpells()
	if ( Grapple.graph ~= nil ) then
		Grapple.graph:Hide();
	end
    GrappleDB.currentFight.book:Hide();
	Grapple.tabGroup:ReleaseChildren();
	local debuffs = GrappleDB.currentFight.debuffs or {};
	local aggros = GrappleDB.currentFight.aggros or {};
    local book = GrappleDB.currentFight.book;
	if ( GrappleDB.fightShown ~= nil and GrappleDB.fightShown > 0 ) then 
		debuffs = GrappleDB.fightHistory[GrappleDB.fightShown].debuffs or {};
		aggros = GrappleDB.fightHistory[GrappleDB.fightShown].aggros or {};
        book = GrappleDB.fightHistory[GrappleDB.fightShown].book;
	end
    if ( GrappleDB.combatBook ~= nil ) then
        GrappleDB.combatBook:Hide();
    end
    Spreadsheet:Load(book);
    GrappleDB.combatBook = book;    
    local damageDone = book:Select("Damage done");
    local damageReceived = book:Select("Damage received");
    damageReceived:HideColumn("NPC", "Spell ID", "Hash");
    damageDone:HideColumn("NPC", "Hash");
	local tbl = {
		{
		value="a",
		text=L["Debuffs applied"],
		children = {}
		},
		{
		value="b",
		text=L["Damage to players by spell"],
        children = {}
		},
        {
		value="c",
		text=L["Damage to players by mob"],
        children = {}
		},
		{
		value="d",
		text=L["Damage to mobs by players"],
        children = {}
		},
        {
		value="$$DamageDoneByFriends=0$",
		text=L["Friendly fire"],
		},
		{
		value="$$Aggro=0$",
		text=L["Bad aggro management"],
		},
	};
	for k, v in pairs(debuffs) do
		local foundSpell = false;
		for sid, stbl in pairs(tbl[1].children) do
			if ( stbl.value == "$$Debuff=" ..v[2] .. "$" ) then
				foundSpell = true;
			end
		end
		if ( foundSpell == false ) then
			local name, rank, icon = GetSpellInfo(v[2]);
			if ( Grapple.goodDebuffs[v[2]] ) then
				table.insert(tbl[1].children, {value="$$Debuff=" ..v[2] .. "$", text="|cff55DD11" .. name .. "|r", icon=icon});
			else
				table.insert(tbl[1].children, {value="$$Debuff=" ..v[2] .. "$", text="|cffDD4400" .. name .. "|r", icon=icon});
			end
		end
	end
	for k, id in pairs (damageReceived:GetUniqueValues("Spell ID")) do
        local name, rank, icon = GetSpellInfo(id);
        table.insert(tbl[2].children, {value="$$DamageReceivedBySpell=" ..id .. "$", text=name, icon=icon});
	end
    for k, id in pairs (damageReceived:GetUniqueValues("NPC")) do
        table.insert(tbl[3].children, {value="$$DamageReceivedByMob=" ..id .. "$", text=id});
	end
    for k, id in pairs (damageDone:GetUniqueValues("NPC")) do
        if ( id ~= "$FF$" ) then
            table.insert(tbl[4].children, {value="$$DamageDoneByMob=" ..id .. "$", text=id});
        end
	end
    
	
	Grapple.tree = AceGUI:Create("TreeGroup");
	Grapple.tree:SetLayout("Flow");
	Grapple.tree:SetCallback("OnClick", 
		function(caller, event, value)
            local subtree, subval = (value):match("%$(%a+)=(.+)%$");
			if ( IsShiftKeyDown() ) then
				if ( ChatFrame1EditBox:IsVisible() ) then
					if ( tonumber(subval) or 0 ~= 0 ) then
						local spellID = tonumber(subval);
						local name, rank, icon = GetSpellInfo(spellID);
						local txt = ChatFrame1EditBox:GetText();
						txt = txt .. GetSpellLink(spellID);
						ChatFrame1EditBox:SetText(txt);
						return;
					end
				end
			end
			local rawList = {};
			if ( subtree == "Debuff" ) then
				local spellID = tonumber(subval);
				Grapple.tree:ReleaseChildren();
                book:Hide();
				local name, rank, icon = GetSpellInfo(spellID);
				table.insert(rawList, (L["Grapple - total applications of debuff %s:"]):format(GetSpellLink(spellID)));
				local stacks = {};
				local maxstacks = 1;
				for k, v in pairs(debuffs) do
					if ( spellID == v[2] ) then
						local found = false;
						for key, char in pairs(stacks) do
							if ( char.guid == v[1] ) then
								found = true;
								char.stacks = char.stacks + 1;
								if ( (v[3] or 0) == 1 ) then
									if ( char.stacktmp > char.stackmax ) then
										char.stackmax = char.stacktmp;
									end
									table.insert(char.stackCount,char.stacktmp);
									char.stacktmp = 0;
								end
								char.stacktmp = char.stacktmp + 1;
								if ( char.stacks > maxstacks ) then maxstacks = char.stacks; end
								break;
							end
						end
						if ( found == false ) then
							local charEntry = GrappleDB.characters:Fetch(v[1]);
							if ( charEntry ~= nil ) then
								table.insert (stacks, {guid=v[1], name=charEntry.name, stacks=1, stacktmp=1, stackmax=1, stackCount = {}});
							end
						end
					end
				end
				table.sort(stacks,
					function (a, b)
					if (a.stacks or 0) > (b.stacks or 0) then
						return true
					elseif (a.stacks or 0) == (b.stacks or 0) then
						if (max(a.stackmax, a.stacktmp)) > (max(b.stackmax, b.stacktmp))  then
							return true
						end
					end
					return false
					end
					);
				local minstacks = 0;
				if ( table.maxn(stacks) > 0 ) then
					minstacks = stacks[table.maxn(stacks)].stacks;
				end
				local y = 0;
				local sheet = Grapple.fumbulator:Select("Debuffs");
                sheet:ClearCells();
                sheet:SetColumn("C", {minval=minstacks, maxval=maxstacks});
                sheet:SetColorRange("C", {0,1,0}, {1,1,0}, {1,0,0});
				for k, v in pairs(stacks) do
					y = y + 1;
					if ( v.stacktmp < 12 ) then
						table.insert(v.stackCount,v.stacktmp);
					end
					sheet:AddRow(y, v.name, v.stacks, table.concat(v.stackCount, "+", 1, x));
					table.insert(rawList, string.format(L["%02u %s: %u stack(s) (%s)"], y, v.name, v.stacks, table.concat(v.stackCount, "+", 1, x)));
				end
				
                Grapple.tree:ReleaseChildren();
                local group = AceGUI:Create("SimpleGroup");
				group:SetHeight(140);
				group.content:SetHeight(140);
				group.width = "fill";
                group:SetAutoAdjustHeight(false);
                Grapple.tree:AddChild(group);
				local btn = AceGUI:Create("Button");
				btn:SetText(L["Export data"]);
				btn:SetWidth(120);
				btn:SetCallback("OnClick", function() Grapple:DataBox(table.concat(rawList, "|n"), rawList, true); end)
				Grapple.tree:AddChild(btn);
                local sframe = Grapple.fumbulator:Render();
                sframe:SetParent(group.content);
                sframe:SetPoint("TOPLEFT", group.content, "TOPLEFT", 0,0);
                sframe:SetPoint("BOTTOMRIGHT", group.content, "BOTTOMRIGHT", -20,0);
                sframe:Show();
			end
			if ( string.sub(subtree or "", 1, 6) == "Damage" ) then
                local f = L["$0. $B: $C hits ($D damage total)"];
                rawList = {};
                Grapple.fumbulator:Hide();
                if ( subtree == "DamageReceivedBySpell" ) then
                    table.insert(rawList, (L["Grapple - total damage received from %s:"]):format(GetSpellLink(tonumber(subval))));
                    GrappleDB.combatBook:Select("Damage received");
                    damageReceived:GroupBy();
                    damageReceived:ClearFilters();
                    damageReceived:AddFilter( { "Spell ID", "=", tonumber(subval)} );
                    damageReceived:GroupBy(L["Player"]);
                    damageReceived:SortByColumn("D", Spreadsheet.sortDescending);
                end
                if ( subtree == "DamageReceivedByMob" ) then
                    table.insert(rawList, (L["Grapple - total damage received from %s:"]):format(subval));
                    GrappleDB.combatBook:Select("Damage received");
                    damageReceived:GroupBy();
                    damageReceived:ClearFilters();
                    damageReceived:AddFilter( { "NPC", "=", subval} );
                    damageReceived:GroupBy(L["Player"]);
                    damageReceived:SortByColumn("D", Spreadsheet.sortDescending);
                    
                end
                if ( subtree == "DamageDoneByMob" ) then
                    local mob = subval;
                    table.insert(rawList, (L["Grapple - total damage done to %s:"]):format(mob));
                    GrappleDB.combatBook:Select("Damage done");
                    damageDone:GroupBy();
                    damageDone:ClearFilters();
                    damageDone:AddFilter( { "NPC", "=", subval} );
                    damageDone:GroupBy(L["Player"]);
                    damageDone:SortByColumn("C", Spreadsheet.sortDescending);
                    f = L["$0. $A: $B hits ($C damage total)"];
                end
                if ( subtree == "DamageDoneByFriends" ) then
                    local mob = L["players (friendly fire)"];
                    table.insert(rawList, (L["Grapple - total damage done to %s:"]):format(mob));
                    GrappleDB.combatBook:Select("Damage done");
                    damageDone:GroupBy();
                    damageDone:ClearFilters();
                    damageDone:AddFilter( {"NPC", "=", "$FF$"} );
                    damageDone:GroupBy(L["Player"]);
                    damageDone:SortByColumn("C", Spreadsheet.sortDescending);
                    f = L["$0. $A: $B hits ($C damage total)"];
                end
                
                GrappleDB.combatBook.selectedSheet.isInteractive = true;
                Grapple.tree:ReleaseChildren();
                local group = AceGUI:Create("SimpleGroup");
				group:SetHeight(140);
				group.content:SetHeight(140);
				group.width = "fill";
                group:SetAutoAdjustHeight(false);
				Grapple.tree:AddChild(group);
                local sframe = GrappleDB.combatBook:Render();
                sframe:SetParent(group.content);
                sframe:SetPoint("TOPLEFT", group.content, "TOPLEFT", 0,0);
                sframe:SetPoint("BOTTOMRIGHT", group.content, "BOTTOMRIGHT", -20,0);
                sframe:Show();
                local btn = AceGUI:Create("Button");
				btn:SetText(L["Export data"]);
				btn:SetWidth(120);
				btn:SetCallback("OnClick", function() 
                    local tbl = GrappleDB.combatBook.selectedSheet:Export(f, true);
                    local newList = {};
                    table.insert(newList, rawList[1]);
                    for k, v in pairs(tbl) do table.insert(newList, v); end
                    Grapple:DataBox(table.concat(newList, "|n"), newList, true);
                    end
                )
				Grapple.tree:AddChild(btn);
            end
                
            
			if ( value == "" ) then
				Grapple.tree:ReleaseChildren();
				local label = AceGUI:Create("Label");
				label:SetWidth(370);
				label:SetText(L["Nothing here yet, but it's coming!"]);
				Grapple.tree:AddChild(label);
			end
			
			if ( subtree == "Aggro" ) then
                book:Hide();
				Grapple.tree:ReleaseChildren();
				table.insert(rawList, (L["Grapple - bad aggro management:"]):format(name));
				local stacks = {};
				local maxstacks = 1;
				for k, v in pairs(aggros) do
					local found = false;
					for key, char in pairs(stacks) do
						if ( char.guid == v[1] ) then
							found = true;
							char.stacks = char.stacks + 1;
							if ( v[2] == 1 ) then char.ones = char.ones + 1;
							else char.twos = char.twos + 1; end
							if ( char.stacks > maxstacks ) then maxstacks = char.stacks; end
							break;
						end
					end
					if ( found == false ) then
						local charEntry = GrappleDB.characters:Fetch(v[1]);
						if ( charEntry ~= nil ) then
							if ( v[2] == 2 ) then  table.insert (stacks, {guid=v[1], name=charEntry.name, stacks=1, ones=0, twos=1});
							else table.insert (stacks, {guid=v[1], name=charEntry.name, stacks=1, ones=1, twos=0});
							end
						end
					end
				end
				table.sort(stacks,
					function (a, b)
					if (a.stacks or 0) > (b.stacks or 0) then
						return true
					elseif (a.stacks or 0) == (b.stacks or 0) then
						if (a.name or "") < (b.name or "")  then
							return true
						end
					end
					return false
					end
					);
				local minstacks = 1;
				local y = 0;
				local sheet = Grapple.fumbulator:Select("Aggro");
                sheet:ClearCells();
                sheet:SetColumn("C", {minval=minstacks, maxval=maxstacks});
                sheet:SetColorRange("C", {0,1,0}, {1,1,0}, {1,0,0});
				for k, v in pairs(stacks) do
					y = y + 1;
					sheet:AddRow(
						y,
						v.name,
						v.ones,
						v.twos
						);
					table.insert(rawList, string.format("%02u %s: %u + %u", y, v.name, v.ones, v.twos));
				end
				local group = AceGUI:Create("SimpleGroup");
				group:SetHeight(140);
				group.content:SetHeight(140);
				group.width = "fill";
				group:SetLayout("Fill");
				group:SetAutoAdjustHeight(false);
				Grapple.tree:AddChild(group);
				local btn = AceGUI:Create("Button");
				btn:SetText(L["Export data"]);
				btn:SetWidth(120);
				btn:SetCallback("OnClick", function() 
                    
                    Grapple:DataBox(table.concat(rawList, "|n"), rawList);
                    end
                );
				Grapple.tree:AddChild(btn);
                local sframe = Grapple.fumbulator:Render();
                sframe:SetParent(group.content);
                sframe:SetPoint("TOPLEFT", group.content, "TOPLEFT", 0,0);
                sframe:SetPoint("BOTTOMRIGHT", group.content, "BOTTOMRIGHT", -20,0);
                sframe:Show();
			end
		end
	);
	Grapple.tree:SetCallback("OnButtonEnter", 
		function(caller, event, value)
			local subtree, subval = (value):match("(%a+)=(%d+)");
			if ( tonumber(subval) or 0 ~= 0 ) then
				local spellID = tonumber(subval);
				GameTooltip:SetOwner(caller.frame, "ANCHOR_TOPLEFT");
				GameTooltip:SetHyperlink("spell:" .. spellID);
				if ( Grapple.goodDebuffs[spellID] ) then
					GameTooltip:AddLine("|cff55DD11[Spell ID: " .. spellID .. "]|r");
				else
					GameTooltip:AddLine("|cffDD4400[Spell ID: " .. spellID .. "]|r");
				end
				GameTooltip:Show();
			end
		end);
	Grapple.tree:EnableButtonTooltips(false);
	Grapple.tree:SetCallback("OnButtonLeave", function() GameTooltip:Hide(); end);
	Grapple.tree:SetLayout("Flow");
	Grapple.tree.width = "fill";
	Grapple.tree.height = "fill";
	Grapple.tree:SetTree(tbl);
	Grapple.tree:SetUserData("tree",tbl)
	Grapple.tabGroup:AddChild(Grapple.tree);
	local label = AceGUI:Create("Label");
	label:SetWidth(370);
	label:SetText(L["The Fumbulator™ shows a wide variety of events that happened to raid players in the selected fight, in case you want to do some naming-and-shaming. Please note, that not all events shown are bad ones. Intentional debuffs and such are included for the sake of properly analysing a fight."]);
	Grapple.tree:AddChild(label);
	
	


end
function Grapple:ShowAbout()
	if ( Grapple.graph ~= nil ) then
		Grapple.graph:Hide();
	end
	local label = AceGUI:Create("Heading");
	label:SetText(string.format(L["About Grapple, revision %u:"], Grapple.revision));
	label:SetWidth(500);
	Grapple.tabGroup:AddChild(label);
	label.label:SetJustifyH("LEFT");
	local label = AceGUI:Create("Label");
    if ( GetLocale() == "zhCN" or GetLocale() == "zhTW" ) then
        label:SetFontObject(LibSpreadsheet_ChineseHeaderFont);
    end
	label.label:SetJustifyH("LEFT");
	label:SetText("Built by: Humbedooh @ Runetotem-EU|n|nThanks for useful feedback goes out to:|n|cffEE9900From <Nightfall> @ Runetotem-EU:|r Alaan, Ardam, Epics, Kitiania, Panora, Tyrygosa|n|cffEE9900From <Rapture> @ Runetotem-EU:|r Thalissa|n|cffEE9900From <Order of the SilverFlame> @ Runetotem-EU:|r Bustah|n|cffEE9900From Curseforge:|r Nevcairiel, Xinhuan, Torhal, Nickenyfiken, Zidomo, OrionShock, Ameyaselene, 羽月_达苏纳|n|cffEE9900From Curse.com:|r Carnacki, Froidre|n|n|cffEE9900Translators:|r Onevcat (Chinese)|n|cffEE9900Libraries used:|r Ace-3.0, LibStub-1.0, LibDataBroker-1.1, LibSpreadsheet-1.0. Thanks to the respective authors!");
	label:SetWidth(500);
    
	Grapple.tabGroup:AddChild(label);
	
end


function Grapple:ShowDetails(what, fight)
	Grapple.tabGroup:ReleaseChildren();
	local label = AceGUI:Create("Label");
	label:SetFontObject(ChatFontLarge);
	label:SetText(L["Yeah, so...that's not gonna happen! Showing the entire session in details would freeze your computer and hurt gnomes. Choose a specific fight instead."]);
	label:SetWidth(500);
	if ( Grapple.skulls == nil ) then
		Grapple.skulls = {};
	end
	for k, v in pairs (Grapple.skulls) do
		v.texture:SetParent(v);
		v:Hide();
		v:SetParent(nil);
	end
	wipe(Grapple.skulls);
	
	if ( Grapple.graph == nil ) then
		Grapple.graph = Graph:CreateGraphLine("Grapple_Realtime_main",Grapple.tabGroup.frame,"BOTTOM","BOTTOM",0,30,480,100)
		Grapple.graph:SetAutoScale(false)
		Grapple.graph:SetGridSpacing(10,50)
		Grapple.graph:SetXAxis(-0, 100)
		Grapple.graph:SetYAxis(-0, 110);
		Grapple.graph.TimeRadius = 10;
		Grapple.graph.Window=Grapple.tabGroup.frame
	end
	Grapple.graph:Hide();
	
	if ( type(what) == "table" ) then
		if ( what.name ~= nil ) then	
			Grapple.tabGroup:AddChild(label);
		else
			local char = GrappleDB.characters:Fetch(what.guid);
			if ( char ~= nil ) then
				char.build = char.build or 0;
				char.class = char.class or "WARRIOR";
				label:SetText(string.format(L["Details about %s's performance:"], char.name));
				Grapple.tabGroup:AddChild(label);
				local btn = AceGUI:Create("Button");
				btn:SetText(L["Refresh"]);
				btn:SetWidth(72);
				btn:SetCallback("OnClick", function() Grapple:ShowDetails(what, fight); end);
				Grapple.tabGroup:AddChild(btn);
				
				local label = AceGUI:Create("Label");
				label:SetText(" ");
				label:SetWidth(12);
				Grapple.tabGroup:AddChild(label);
				
				label = AceGUI:Create("ColorPicker");
				label:SetLabel(L["GrappleScore"]);
				label:SetColor(1,0.95,0.1,1);
				label:SetWidth(110);
				label:SetDisabled(true);
				Grapple.tabGroup:AddChild(label);
				label = AceGUI:Create("ColorPicker");
				label:SetLabel(L["Activity"]);
				label:SetColor(0,0.8,0,1);
				label:SetWidth(75);
				label:SetDisabled(true);
				Grapple.tabGroup:AddChild(label);
				label = AceGUI:Create("ColorPicker");
				label:SetLabel(L["Survivability"]);
				label:SetColor(0.95,0.25,0.1,1);
				label:SetWidth(100);
				label:SetDisabled(true);
				Grapple.tabGroup:AddChild(label);
				label = AceGUI:Create("ColorPicker");
				label:SetLabel(L["Health"]);
				label:SetColor(0.8,0.1,0.55,1);
				label:SetWidth(70);
				label:SetDisabled(true);
				Grapple.tabGroup:AddChild(label);
				label = AceGUI:Create("ColorPicker");
				label:SetLabel(L["Power"]);
				label:SetColor(0.1,0.2,0.8,1);
				label:SetWidth(65);
				label:SetDisabled(true);
				Grapple.tabGroup:AddChild(label);

				if ( what.history ~= nil ) then
					local tmax = table.maxn(what.history) - 1;
					if ( tmax < 1 ) then
						tmax = 1;
					end
					local a = {{0,0}};
					local b = {{0,0}};
					local c = {{0,0}};
					local d = {};
					local e = {};
					local i = 1;
					local duration = fight.raid.uptime + 1;
					if ( duration < 1 ) then	
						duration = 1;
					end
					local y = 0;
					for xk, xv in pairs (what.history or {}) do
						y = y + 1;
					-- order is: fightID, now, damage, activity, uptime, raidEntry, hpHistory
						local charEntry = {}
						charEntry.fightID = xv[1];
						charEntry.now = xv[2];
						charEntry.damage = xv[3];
						charEntry.activity = xv[4];
						charEntry.uptime = xv[5];
						local raidEntry = {};
						raidEntry.fightID = fight.raid.history[y][1];
						raidEntry.now = fight.raid.history[y][2];
						raidEntry.damage = fight.raid.history[y][3];
						raidEntry.activity = fight.raid.history[y][4];
						raidEntry.uptime = fight.raid.history[y][5];
						local tbl = Grapple:Calculate(char.class, what.build, charEntry, raidEntry, fight);
						i = i + raidEntry.uptime;
						local aa = tonumber(( ((tbl.activityScore or 1) * 35) / Grapple.scoreModifiers[char.class][char.build][2])+4);
						local bb = tonumber(((tbl.survivabilityScore or 1) / 2)+6);
						local cc = tonumber(((tbl.grappleScore or 1) / 8 ));
						if ( aa > 105 ) then aa = 105; end
						if ( bb > 105 ) then bb = 105; end
						if ( cc > 105 ) then cc = 105; end
						if ( aa < 1 ) then aa = 1; end
						if ( bb < 1 ) then bb = 1; end
						if ( cc < 1 ) then cc = 1; end
						local x = (i/duration) * 100;
						if ( x > 100 ) then x = 100; end
						table.insert(a, 1, { x, aa });
						table.insert(b, 1, { x, bb });
						table.insert(c, 1, { x, cc });
					end
					local y = table.maxn(what.hpHistory);
					for k, v in pairs ( what.hpHistory ) do
						if ( k == 1 ) then
							table.insert(d, 1, { 0, v  });
						end
						table.insert(d, 1, { ((k)/y) * 100 , v  });
					end
					if ( what.pwHistory ~= nil ) then
						y = table.maxn(what.pwHistory);
						for k, v in pairs ( what.pwHistory ) do
							if ( k == 1 ) then
								table.insert(e, 1, { 0, v  });
							end
							table.insert(e, 1, { ((k)/y) * 100 , v  });
						end
					end
					Grapple.graph.Data = {};
					Grapple.graph.FilledData = {};
					tinsert(Grapple.graph.FilledData,{ Points=d ;Color={0.9,0.1,0.6,0.2}} );
					tinsert(Grapple.graph.FilledData,{ Points=a ;Color={0,0.8,0.2,0.3}} );
					tinsert(Grapple.graph.Data,{ Points=e ;Color={0.1,0.1,0.9,0.95}} );
					tinsert(Grapple.graph.Data,{ Points=d ;Color={0.8,0.1,0.55,0.75}} );
					tinsert(Grapple.graph.Data,{ Points=a ;Color={0,0.8,0,0.85}} );
					tinsert(Grapple.graph.Data,{ Points=b ;Color={0.95,0.25,0.1,1}} );
					tinsert(Grapple.graph.Data,{ Points=c ;Color={1,0.95,0.1,1}} );
					Grapple.graph:RefreshGraph()
					Grapple.graph:Show();
					local swords = {};
					local skulls = {};
					local tombstones = {};
					for k, v in pairs(fight.deaths) do							
						local name = v[3]
						local when = (((v[1] - fight.start - 2) / duration) * 480) - 20;
						if ( when > 460 ) then when = 460; end
						
						local cause = "";
						local causeRaw = {
						[1] = ("%s died at %s"):format(name, date("%H:%M:%S", v[1])),
						};
						if ( v[4] ~= nil ) then
							for lk, lv in pairs (v[4]) do
								cause = lv .. cause;
								table.insert(causeRaw, 1, lv );
							end
						else
							cause = L["|cffffffffLiver failure?|r"];
						end
						local cause = L["Cause:|n"] .. cause;
						
						local xchar = GrappleDB.characters:Fetch(v[2]);
						if ( xchar ~= nil ) then
							if ( char.guid == xchar.guid ) then
								table.insert(tombstones, {when+40, 15, cause, causeRaw, name, v[1]});
							else
								table.insert(skulls, {when+40, 15, cause, causeRaw, name, v[1]});
							end
						else
							table.insert(swords, {when+40, 115, cause, causeRaw, name, v[1]});
						end						
					end
					for k, val in pairs(swords) do
						local f = CreateFrame("Button");
						f:SetWidth(20);
						f:SetHeight(28);
						f.texture = f:CreateTexture("overlay", "OVERLAY");
						f.texture:SetTexture("Interface\\AddOns\\Grapple\\Textures\\sword");
						f:SetPoint("BOTTOMLEFT", Grapple.graph.Window, "BOTTOMLEFT", val[1], val[2]);
						f:RegisterForClicks("RightButtonUp")
						f.texture:SetAllPoints(f);
						f:SetParent(Grapple.graph.TextFrame);
						f:SetScript("OnEnter", function(self) GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT"); GameTooltip:ClearLines() GameTooltip:AddLine( ("%s died at %s"):format(val[5], date("%H:%M:%S", val[6]))); GameTooltip:AddLine(val[3]); GameTooltip:AddLine("(Right click to copy log to a textbox)"); GameTooltip:Show(); end);
						f:SetScript("OnLeave", function(self) GameTooltip:Hide(); end);
						f:SetScript("OnClick", function(self) Grapple:DataBox(("%s died at %s|n"):format(val[5], date("%H:%M:%S", val[6])) .. val[3], val[4]); end);
						f:EnableMouse(true);
						table.insert(Grapple.skulls, f);
					end
					for k, val in pairs(tombstones) do
						local f = CreateFrame("Button");
						f:SetWidth(20);
						f:SetHeight(28);
						f.texture = f:CreateTexture("overlay", "OVERLAY");
						f.texture:SetTexture("Interface\\AddOns\\Grapple\\Textures\\tombstone");
						f:SetPoint("BOTTOMLEFT", Grapple.graph.Window, "BOTTOMLEFT", val[1], val[2]);
						f:RegisterForClicks("RightButtonUp")
						f.texture:SetAllPoints(f);
						f:SetFrameStrata("FULLSCREEN_DIALOG");
						f:SetParent(Grapple.graph.TextFrame);
						f:SetScript("OnEnter", function(self) GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT"); GameTooltip:ClearLines() GameTooltip:AddLine( ("%s died at %s"):format(val[5], date("%H:%M:%S", val[6]))); GameTooltip:AddLine(val[3]); GameTooltip:AddLine("(Right click to copy log to a textbox)"); GameTooltip:Show(); end);
						f:SetScript("OnLeave", function(self) GameTooltip:Hide(); end);
						f:SetScript("OnClick", function(self) Grapple:DataBox(("%s died at %s|n"):format(val[5], date("%H:%M:%S", val[6])) .. val[3], val[4]); end);
						f:EnableMouse(true);
						table.insert(Grapple.skulls, f);
					end
					for k, val in pairs(skulls) do
						local f = CreateFrame("Button");
						f:SetWidth(24);
						f:SetHeight(24);
						f.texture = f:CreateTexture("overlay", "OVERLAY");
						f.texture:SetTexture("Interface\\AddOns\\Grapple\\Textures\\skull");
						f:SetPoint("BOTTOMLEFT", Grapple.graph.Window, "BOTTOMLEFT", val[1], val[2]);
						f:RegisterForClicks("RightButtonUp")
						f.texture:SetAllPoints(f);
						f:SetParent(Grapple.graph.TextFrame);
						f:SetScript("OnEnter", function(self) GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT"); GameTooltip:ClearLines() GameTooltip:AddLine( ("%s died at %s"):format(val[5], date("%H:%M:%S", val[6]))); GameTooltip:AddLine(val[3]); GameTooltip:AddLine("(Right click to copy log to a textbox)"); GameTooltip:Show(); end);
						f:SetScript("OnLeave", function(self) GameTooltip:Hide(); end);
						f:SetScript("OnClick", function(self) Grapple:DataBox(("%s died at %s|n"):format(val[5], date("%H:%M:%S", val[6])) .. val[3], val[4]); end);
						f:EnableMouse(true);
						table.insert(Grapple.skulls, f);
					end
					
				end
			else
				local label = AceGUI:Create("Label");
				label:SetWidth(480);
				label.label:SetJustifyH("LEFT");
				label:SetText(L["Click a player name for details.|n|nThe graph will show up to the last 15 minutes of activity in the selected fight or session.|nThe |cff33EE22green|r filled area represents the raider's spell activity, the |cffEE3322red|r line represents the survival score and the |cffEEEE22yellow|r line represents the GrappleScore.|n|nIf you are monitoring an ongoing fight, use the refresh button to update the graph."]);
				Grapple.tabGroup:AddChild(label);
			end
		end
	else
		local label = AceGUI:Create("Label");
		label:SetWidth(480);
		label.label:SetJustifyH("LEFT");
		label:SetText(L["Click a player name for details.|n|nThe graph will show up to the last 15 minutes of activity in the selected fight or session.|nThe |cff33EE22green|r filled area represents the raider's spell activity, the |cffEE3322red|r line represents the survival score and the |cffEEEE22yellow|r line represents the GrappleScore.|n|nIf you are monitoring an ongoing fight, use the refresh button to update the graph."]);
		Grapple.tabGroup:AddChild(label);
	end
	
end


function Grapple:PopulateWindow()
	if ( Grapple.status ~= nil ) then
		Grapple.status:ReleaseChildren();
	else
		Grapple.status = AceGUI:Create("SimpleGroup");
		Grapple.status:SetLayout("Flow")
		Grapple.status:SetWidth(600);
		Grapple.status:SetHeight(230);
		Grapple.status:SetAutoAdjustHeight(false);
		Grapple.mainWin:AddChild(Grapple.status);
	end
	local rows = 0;
	local rIndex = {};
	
	local caption = AceGUI:Create("Heading");
	local chosenFight = L["Entire session"];
	local history = GrappleDB.fightHistory[GrappleDB.fightShown];
	
	if ( GrappleDB.fightShown == 0 or GrappleDB.fightShown == nil ) then
		history = GrappleDB.currentFight;
		chosenFight = L["Current fight"];
	end
	if ( GrappleDB.fightShown > 0 ) then
		chosenFight = GrappleDB.fightHistory[GrappleDB.fightShown].title;
	end
	Grapple.status:AddChild(caption);
	

	local entire = false;
	if ( history == nil ) then
		history = Grapple:FightList();
		entire = true;
		local uptime = 0;
		for title, xhistory in pairs (GrappleDB.fightHistory) do
			uptime = uptime + xhistory.raid.uptime;
			for k, v in pairs(xhistory.characters.data) do
				local char = GrappleDB.characters:Fetch(v.guid);
				if ( char ~= nil and char.class ~= nil and v.activity > 1) then
					local char = history.characters:Fetch(v.guid);
					if ( char == nil ) then
						char = history.characters:Add(v.guid, "thisaintgonnawork");
						char.build = v.build;
						char.uptime = 0;
					end
					char.focusFumbles = char.focusFumbles + (v.focusFumbles * xhistory.raid.uptime)
					char.survivalFumbles = char.survivalFumbles + (v.survivalFumbles * xhistory.raid.uptime)
					char.grappleScore = char.grappleScore + (v.grappleScore * xhistory.raid.uptime);
					char.activityScore = char.activityScore + (v.activityScore * xhistory.raid.uptime);
					char.uptimeScore = char.uptimeScore + (v.uptimeScore * xhistory.raid.uptime);
					char.focusScore = char.focusScore + (v.focusScore * xhistory.raid.uptime);
					char.survivabilityScore = char.survivabilityScore + (v.survivabilityScore * xhistory.raid.uptime);
					char.uptime = (char.uptime or 0) + xhistory.raid.uptime;
					char.activity = 1;
				end
			end
		end
		if ( Grapple.inCombat == true ) then
			if ( GrappleDB.currentFight.raid.uptime ~= nil ) then -- make sure we have some data first.
				if ( GrappleDB.currentFight.raid.uptime >= 30 ) then
					uptime = uptime + GrappleDB.currentFight.raid.uptime;
					for k, v in pairs(GrappleDB.currentFight.characters.data) do
						local char = GrappleDB.characters:Fetch(v.guid);
						if ( char ~= nil and v.activity > 1) then
							local char = history.characters:Fetch(v.guid);
							if ( char == nil ) then
								char = history.characters:Add(v.guid, "thisaintgonnawork");
								char.build = v.build;
							end
							char.focusFumbles = char.focusFumbles + (v.focusFumbles * GrappleDB.currentFight.raid.uptime)
							char.survivalFumbles = char.survivalFumbles + (v.survivalFumbles * GrappleDB.currentFight.raid.uptime)
							char.grappleScore = char.grappleScore + (v.grappleScore * GrappleDB.currentFight.raid.uptime);
							char.activityScore = char.activityScore + (v.activityScore * GrappleDB.currentFight.raid.uptime);
							char.uptimeScore = char.uptimeScore + (v.uptimeScore * GrappleDB.currentFight.raid.uptime);
							char.focusScore = char.focusScore + (v.focusScore * GrappleDB.currentFight.raid.uptime);
							char.survivabilityScore = char.survivabilityScore + (v.survivabilityScore * GrappleDB.currentFight.raid.uptime);
							char.uptime = (char.uptime or 0) + GrappleDB.currentFight.raid.uptime;
						end
					end
				end
			end
		end
		for k, v in pairs(history.characters.data) do
			v.grappleScore = v.grappleScore / v.uptime;
			v.activityScore = v.activityScore / v.uptime;
			v.uptimeScore = v.uptimeScore / v.uptime;
			v.focusScore = v.focusScore / v.uptime;
			v.survivabilityScore = v.survivabilityScore / v.uptime;
			v.focusFumbles = v.focusFumbles / v.uptime;
			v.survivalFumbles = v.survivalFumbles / v.uptime;
		end
		caption:SetText(string.format(L["Data for %s:"], chosenFight));
	end
    
    local sheet = Grapple.scoreboard:Select("Scores");
    sheet:ClearCells();
	
	if ( history ~= nil ) then
		if ( entire == false ) then
			if ( history.raid == nil ) then
                history.raid = {};
			end
			if ( history.raid.uptime ~= nil ) then
				local s = math.fmod(history.raid.uptime, 60);
				local m = math.floor((math.fmod(history.raid.uptime, 3600) - s)/60);
				chosenFight = chosenFight .. string.format(" (%02d:%02d)", m, s);
			end
			caption:SetWidth(500);
			caption:SetHeight(20);
			caption:SetText(string.format(L["Data for %s:"], chosenFight));

		end
		table.sort(history.characters.data, Grapple.SortFunc);
		for k, v in pairs(history.characters.data) do
			local char = GrappleDB.characters:Fetch(v.guid);
			if ( char ~= nil and v.activity > 0 ) then
                
				rows = rows + 1;
                local x = 0.25;
                if ( v.role == 1 ) then x = 0.50;
                elseif ( 
                    (char.class == "PRIEST" and (v.build == 2 or v.build == 3 )) or
                    (char.class == "PALADIN" and (v.build == 2)) or
                    (char.class == "DRUID" and (v.build == 2)) or
                    (char.class == "SHAMAN" and (v.build == 1))
                    ) then
                    x = 0.75;
                end
				local r = sheet:AddRow(
                        {data=char.name, id={v,history}, color=Grapple.colors[char.class], icon=[[Interface\LFGFrame\LFGRole]], texcoords={x, x+0.25,0,1} },
                        Grapple.scoreModifiers[char.class][(v.build or char.build)][1],
                        v.uptimeScore,
                        v.survivabilityScore,
                        v.activityScore * 60,
                        string.format("%02u + %02u", (v.focusFumbles or 0), (v.survivalFumbles or 0)),
                        v.focusScore,
                        v.grappleScore
                    );
                    --[[sheet:SetRow(r, {background=
                        {
                        (Grapple.colors[char.class].r/6)+0.07,
                        (Grapple.colors[char.class].g/6)+0.07,
                        (Grapple.colors[char.class].b/6)+0.07,
                        1
                        }
                    }
                    );]]
			end
		end
		if ( rows > 0 ) then
			local f = Grapple.scoreboard:Render();
			f:SetParent(Grapple.status.frame);
            f:SetPoint("TOPLEFT", Grapple.status.frame, "TOPLEFT", 5, -20);
            f:SetPoint("BOTTOMRIGHT", Grapple.status.frame, "BOTTOMRIGHT", -20, 20);
            f:Show();
        end
	end
	
	if ( rows == 0 ) then
		label = AceGUI:Create("Label");
		label:SetWidth(500);
		label:SetText(L["No log entries found in fight or session.|n|nData will become available in the 'Current fight' entry whenever a fight starts."]);
		label.label:SetJustifyH("LEFT");
        Grapple.scoreboard:Hide();
		Grapple.status:AddChild(label);
	end
	if ( Grapple.currentTab == "spells" and (GrappleDB.fightShown or 0) == 0 ) then Grapple:ShowSpells(); end
end

function Grapple:SilentPopulate(history)
	if ( history ~= nil ) then
		for k, v in pairs(history.characters.data) do
			local char = GrappleDB.characters:Fetch(v.guid);
			if ( char ~= nil and char.class ~= nil and v.uptime > 0) then
				local tbl = Grapple:Calculate(char.class, (v.build or 0), v, history.raid, history);
				v.grappleScore = tbl.grappleScore;
				v.activityScore = tbl.activityScore;
				v.uptimeScore = tbl.uptimeScore;
				v.focusScore = tbl.focusScore;
				v.survivabilityScore = tbl.survivabilityScore;
			end
		end
	else
		for k, v in pairs(GrappleDB.fightHistory) do
			Grapple:SilentPopulate(v);
		end
		if ( Grapple.mainWin:IsVisible() ) then
			Grapple:SilentPopulate(GrappleDB.currentFight);
			Grapple:PopulateWindow();
		end
	end
end

function Grapple:ShowOptions()
    Grapple.fumbulator:Hide();
	if ( Grapple.graph ~= nil ) then
		Grapple.graph:Hide();
	end
	Grapple.tabGroup:ReleaseChildren();
	local scroller = AceGUI:Create("ScrollFrame");
	scroller:SetLayout("Fill");
	scroller:SetWidth(590);
	scroller:SetHeight(170);
	Grapple.tabGroup:AddChild(scroller);
	local window = AceGUI:Create("SimpleGroup");
	window:SetLayout("Flow")
	window:SetWidth(590);
	window:SetHeight(200);
	scroller:AddChild(window);
	
	local label = AceGUI:Create("Label");
	label:SetWidth(500);
	label:SetText(L["This window allows you to modify some of the internal constants used to calculate the scores.|nApM is the expected Activity per Minute and Surv is the expected mortality rate.|nEdit at your own risk!|nTo save a value, simply enter it and press Enter - the scoreboard will update immediately after."]);
	window:AddChild(label);
	
	for k, v in pairs(Grapple.scoreModifiers) do
		local label = AceGUI:Create("Heading");
		label:SetWidth(500);
		label:SetText(k);
		window:AddChild(label);
		for ck, cv in pairs(v) do
			local label = AceGUI:Create("Label");
			label:SetWidth(55);
			local caption = cv[1];
			if ( string.len(caption) > 8 ) then
				caption = string.sub(caption, 0, 6) .. "..";
			end	
			label:SetText(caption .. ":");
			window:AddChild(label);
			local box = AceGUI:Create("EditBox");
			box:SetWidth(25);
			if ( box.DisableButton ~= nil ) then box:DisableButton(true); end
			box:SetLabel("ApM");
			box:SetText(cv[2]*60);
			box.showbutton = false;
			box:SetCallback("OnEnterPressed", function(caller, event, val) cv[2] = tonumber(val) / 60; Grapple:SilentPopulate(nil); end);
			window:AddChild(box);
			label = AceGUI:Create("Label");
			label:SetWidth(5);
			label:SetText("");
			window:AddChild(label);
			box = AceGUI:Create("EditBox");
			box:SetWidth(35);
			box:SetLabel("Surv");
			box:SetText(cv[3]*100);
			box.showbutton = false;
			box:SetCallback("OnEnterPressed", function(caller, event, val) cv[3] = tonumber(val) / 100; Grapple:SilentPopulate(nil); end);
			window:AddChild(box);
			label = AceGUI:Create("Label");
			label:SetWidth(5);
			label:SetText("");
			window:AddChild(label);
		end
	end
	local label = AceGUI:Create("Heading");
	label:SetWidth(500);
	label:SetText("Math functions");
	window:AddChild(label);
	
	scroller.content:SetHeight(1200);
	
end


function Grapple:DataBox(text, raw, noStrip) 
	Grapple.textboxFrame = AceGUI:Create("Frame");
	Grapple.textboxFrame:SetWidth(540);
	Grapple.textboxFrame:SetHeight(300);
	Grapple.textboxFrame:SetLayout("Flow");
	Grapple.textboxFrame:SetTitle(L["Data output"]);
	Grapple.textbox = AceGUI:Create("MultiLineEditBox");
	Grapple.textbox:SetWidth(530);
	Grapple.textbox:SetHeight(170);
	Grapple.textboxFrame:AddChild(Grapple.textbox);
	Grapple.textbox:SetText(text);
	
	GrappleDB.dataRecipient = 0;
	drop = AceGUI:Create("Dropdown");
	drop:SetWidth(140);
	drop:SetLabel(L["Broadcast channel:"]);
	drop:SetCallback("OnValueChanged", function(key) GrappleDB.dataRecipient = key.value; end);
	local tbl = deepcopy(Grapple.reportList);
	table.remove(tbl, 7);
	for n = 1, 9 do
		local id, name = GetChannelName(n);
		if ( id ~= 0 ) then
			tbl[id + 10] = id .. ". " .. name;
		end
	end
	drop:SetList(tbl);
	drop:SetText(tbl[GrappleDB.dataRecipient]);
	Grapple.textboxFrame:AddChild(drop);
	
	drop = AceGUI:Create("Dropdown");
	drop:SetWidth(100);
	drop:SetLabel(L["Max no. of entries:"]);
	local tbl = {};
	for i = 1, 25 do
		table.insert(tbl, i);
	end
	table.insert(tbl, 40);
	GrappleDB.recipientLimit = 10;
	drop:SetList(tbl);
	drop:SetText("10");
	drop:SetCallback("OnValueChanged", function(key) GrappleDB.recipientLimit = key.value; end);
	Grapple.textboxFrame:AddChild(drop);

	
	local btn = AceGUI:Create("Button");
	btn:SetText(L["Send"]);
	btn:SetWidth(60);
	btn:SetCallback("OnClick", function()
		local where = Grapple.reportList[GrappleDB.dataRecipient];
		local channel = 0;
		if ( GrappleDB.dataRecipient > 10 ) then
			where = "Channel";
			channel = GrappleDB.dataRecipient - 10;
		end
		local i = 0;
		for k, line in pairs(raw) do
			if ( i > GrappleDB.recipientLimit ) then break; end
			local xline = string.gsub(line, "|cff%w%w%w%w%w%w", ""); -- remove formatting so we won't get kicked off the server.
			xline = string.gsub(xline, "|%w", ""); 
			if ( noStrip ~= nil ) then
				xline = line;
			end
			SendChatMessage(xline, where, nil, channel);
			i = i + 1;
		end
	end);
	Grapple.textboxFrame:AddChild(btn);
	
	Grapple.textboxFrame:Show();
end


function Grapple:MessageBox(text) 
	Grapple.textboxFrame = AceGUI:Create("Frame");
	Grapple.textboxFrame:SetWidth(400);
	Grapple.textboxFrame:SetHeight(200);
	Grapple.textboxFrame:SetLayout("Fill");
	Grapple.textboxFrame:SetTitle(L["Grapple"]);
	Grapple.textbox = AceGUI:Create("Label");
	Grapple.textboxFrame:AddChild(Grapple.textbox);
	Grapple.textbox:SetText(text);
	Grapple.textboxFrame:Show();
end


function Grapple:BroadcastList()
	local obj = AceGUI:Create("SimpleGroup");
	obj:SetLayout("Flow");
	obj.width = "fill";
	obj:SetAutoAdjustHeight(false);
	obj:SetHeight(20);
	local drop = AceGUI:Create("Dropdown");
	drop:SetWidth(140);
	drop:SetLabel(L["Broadcast channel:"]);
	drop:SetCallback("OnValueChanged", function(key) GrappleDB.dataRecipient = key.value; end);
	local tbl = deepcopy(Grapple.reportList);
	table.remove(tbl, 7);
	for n = 1, 9 do
		local id, name = GetChannelName(n);
		if ( id ~= 0 ) then
			tbl[id + 10] = id .. ". " .. name;
		end
	end
	drop:SetList(tbl);
	drop:SetText(tbl[GrappleDB.dataRecipient]);
	obj:AddChild(drop);
	
	drop = AceGUI:Create("Dropdown");
	drop:SetWidth(100);
	drop:SetLabel(L["Max no. of entries:"]);
	local tbl = {};
	for i = 1, 25 do
		table.insert(tbl, i);
	end
	table.insert(tbl, 40);
	GrappleDB.recipientLimit = 25;
	drop:SetList(tbl);
	drop:SetText("25");
	drop:SetCallback("OnValueChanged", function(key) GrappleDB.recipientLimit = key.value; end);
	obj:AddChild(drop);
	
	local btn = AceGUI:Create("Button");
	btn:SetText(L["Send report"]);
	btn:SetWidth(100);
	btn:SetCallback("OnClick", function() Grapple:Report(Grapple.reportList[GrappleDB.recipient]); end)
	obj:AddChild(btn);
	return obj;
end